#
# Python implementation of
# Searching Seed Code Sequences for generating Single Track Gray Codes
#
# Words don't come easy to me
#
#
# For given n generate all Lyndon Words that are Necklaces
# In this set search for cyclic sequences where following workds differ 
# only in one bit  "Snakes" 
#
# Invokation:
# python3 lyndon_words.py -m<method> <n>  [<seqlen>]
# where n=      number of Bits
# and   seqlen= length of sequence
#
#
# Method to be used to be uncommented at the end of this program
# Copy output sequence and use it for stgc_process.py
#
# 
# siehe auch https://gist.github.com/dvberkel/1950267

import random 
import sys
import graph
from operator import itemgetter


n =5 # Max recursion < 14
# must be adjusted accordingly

if len(sys.argv) > 1:
  n=int(sys.argv[1])
else:
  n=1
seqlen=3
if len(sys.argv) > 2:
  seqlen=int(sys.argv[2])
else:
  seqlen=3
  
#S = ['0', '1', '2']
S = ['0', '1']
Tc=0

k = len(S)
S.sort()
lyndonwords = []
Adjm = []

#
# gray transition of two words 
#  position where bit flips
#
def transition(word,word2):
   for i in range(0,len(word)):
      c1=word[i]
      c2=word2[i]
      if word[i] != word2[i]:
         return i
   return None

#
# apply a bit flip of the i-th bit to a word
#
def applyTransition(word,i):
  #print("a",word,i)
  if word[i]=='0':
     return word[0:i] + '1' + word[i+1:]
  else:
     return word[0:i] + '0' + word[i+1:]
  

#
# apply a sequence of transitions to word
#
def applyTransitionSequence(word,tseq,words):
   words.append(word)
   for t in tseq:
      word=applyTransition(word,t)
      words.append(word)
   return words
   
   
#
# get the transition sequence of word list:
#
def transitionSequence(words):
  ts=[]
  for i in range(0,len(words)):
    ts.append( transition( words[i],words[(i+1) % len(words)]))
  return ts
    

#
# normalize the transition sequence
#
def normalizeSequence(tseq):
  k = 0
  t = {}
  ntseq = []  
  for i in range(0,len(tseq)):
     if tseq[i] in t.keys():
        ntseq.append( t[tseq[i]] )
     else:
        ntseq.append(k)
        t[tseq[i]]=k
        k=k+1
  return ntseq
   
#print( normalizeSequence([1,0,3,2,1,0,3,2]))
  

#
# initialized adjacence matrix
#
def initAdjm(n):
   for i in range(0,n):
      Adjm.append([])
      for j in range(0,n):
         Adjm[i].append(0)
#
# print adjacence matrix
#
def printAdjm(n):
   for i in range(0,n):
      print(Adjm[i])
    
      
# cyclic left shift of word
def e(word):
   sword=''
   for i in range(1,len(word)):
     sword=sword + word[i]
   sword = sword + word[0]
   return sword 

# l times cyclic left shift of word
def el(word,l):
  sword = word
  for i in range(0,l):
     sword = e(sword)   
  return sword
  
#  check if word is a lyndon word
#  smallest lexicographic rotation of a word
#
def isLyndon(w):
  flag=True
  for i in range(1,len(w)):
    #print(w,el(w,i))
    if w > el(w,i): 
       flag=False
  return flag
 
#  check if word is a necklace
#
def isNecklace(w):
   flag=True
   for l in range(1,len(w)):
     if w==el(w,l):
        flag=False
        break
   return flag
   

#
# test if w1 and w2 have gray distance
#
def gray(w1,w2):
   d=0
   for i in range( 0, len(w1) ):
     if w1[i]!=w2[i]:
       d=d+1
       if d>1:
           break
   if d==1:
       return True
   else:
       return False   
      
 

 #
 # fill adnacence matrix
 #
def fillAdjm(words,n):
     edges=0
     for i in range(0,n):
      for j in range(0,n):
         if(gray(words[i],words[j])):
            edges = edges+1
            Adjm[i][j]=1
     print("Edges:" ,edges)


         
def sort_gray(l,words,acc):
  # print(l,words,acc)
   if words==[]:
      print("Result is ", len(acc),acc)
   else:
      w1=acc[-1]
     # print(w1,words)
      rest = words[0::]
      for w2 in rest:
        # print(w1,w2, gray(w1,w2))
         if gray(w1,w2):
             rest2 = rest.copy()
             rest2.remove(w2)
             acc2= acc.copy()
             acc2.append(w2)
             #print("Hure ",rest2, acc2)
             sort_gray(l+1,rest2,acc2)

neighbors = []

def split_min_max_1(ns):
   m=None
   mi=[]
   ma=[]
   lns2=list(map(lambda i: len(neighbors[i]), ns)) 
   m=min(lns2)
   for n in range(0,len(ns)):
      if lns2[n]==m:
           mi.append(n)
      else:
           ma.append(n)
      #print(n,lns2[n])     
           
   return (mi,ma)
   
def split_min_max():
   m=None
   mi=[]
   ma=[]
   for n in range(0,len(neighbors)):
      if m==None:
         m = len(neighbors[n])
      else:
         if m > len(neighbors[n]):
            m=len(neighbors[n])
   for n in range(0,len(neighbors)):
      if len(neighbors[n])==m:
           mi.append(n)
      else:
           ma.append(n)
   #print(m)
   return (mi,ma)
           
            
def find_neighbors(nodes):
  for n in range(0,len(nodes)):
     neighbors.append([])
     for j in range(0,len(nodes)):
         if Adjm[n][j]==1:
            neighbors[n].append(j)
            print(" ({},{} -> {})".format(n,j,neighbors[n]))
  for n in range(0,len(nodes)):
      print("{} {}: {} {}".format(n,nodes[n],neighbors[n], list(map(lambda i: nodes[i], neighbors[n]))))  
  print(neighbors)
  
# number of differences of two words
#
def differences(w1,w2):
  d=0
  for i in range(0,len(w1)):
    if w2[i] != w1[i] :
       d = d+1
  return d
  
# check if word w is compatible with snake 
#
#
def isCompatible(w,snake):
   #print(w,snake)
   if snake == []:
     return True
   if len(snake)==1:
      if differences(w,snake[0])==1:
         return True
   for i in range(1,len(snake)-2):
     if differences(w,snake[i])==1 :
       #  print("Not compatible: ",w, snake)
         return False
   if differences(w,snake[len(snake)-1])==1:
      return True
   if differences(w,snake[0])==1:
      return True
 
   return False   
   
# affe = [("0000011", ['0000101', '0001101', '0001111', '0011111', '0010111', '0010011', '0000011', '0000001']),
#("0000101" ,  ['0000101', '0001101', '0001111', '0011111', '0010111', '0010011', '0000011', '0000001']),
#("0001001" ,  ['0000101', '0001101', '0001111', '0011111', '0010111', '0010011', '0000011', '0000001'])]
#  
#print(affe)
#for (k,l) in affe:
#    print( isCompatible(k,l))

#o=8/0     


mins=[]
maxs=[]
global mx
mx=0

#
# recursive search for cycles
# level    :
# i        :
# acc      :
# visited  :
# m        :
# start    :
# neighbors:
#
def find_snaker(level,i,acc,visited,m,start,neighbors):
   #print(level)
  # print("N",neighbors)
   cycle=None
   found=False
   global mx
                 
   if i==None:
       for i in range(0,m):
         acc.append(i)
         visited[i]=1
         find_snaker(level+1,i,acc,visited,m,i,neighbors)
         found=True
         visited[i]=0
         acc.remove(i)
   else:   
    
      v= neighbors[i].copy()
   
     # Remove Neighbors that are already in sequence
      for a in acc:
         if a in v:
           v.remove(a)
           
           
      # find compatible nodes   
      # "real Snakes": no loops to inner nodes within acc:
      #vc = []
      #for j in v:
      #   if isCompatible(lyndonwords[j],list(map(lambda x: lyndonwords[x], acc))):
      #     vc.append(j)
      #print("V VC",v,vc,list(map(lambda x: lyndonwords[x], acc)))
      vc=v
      
      # Count Neighbours of all Neigbours in vc
      #vcn = []
      #mi_vc = []
      #ma_vc = []
      #mins=1000000000
      #for i in vc:
      #    ns=neighbors[i]
      #    if len(ns ) < mins:
      #      mins=len(ns)
      #    vcn.append( len(ns) )
          
      #for i in range(0,len(vc)):
      #    if vcn[i] == mins:
      #      mi_vc.append(vc[i])
      #    else:
      #      ma_vc.append(vc[i])
      #  
      #  
    # # print("MINS",mi_vc)   
    #  list(map(lambda x: print(x,len(x)), mi_vc))
    #  print("MAXS",ma_vc)
    #  list(map(lambda x: print(x,len(x)), ma_vc))
    
     # with 50% chance take neighbour with least neigbours
      #r=random.sample([True,False],k=1)
     
      #if r:
      #    random.shuffle(mi_vc)
      #    random.shuffle(ma_vc)
      #    vc=mi_vc + ma_vc
      #else:
      #   if not vc==[]:
      #      random.shuffle(vc)
            
  #    print("VC",level,acc,v,vc)
   #   if not acc == []:
   #      k=acc[-1]
   #      print(k,neighbors[k],neighbors)
      
      if not vc==[]:
         random.shuffle(vc)
      
        
      for j in vc:
         #   print(lyndonwords[j],acc,list(map(lambda x: lyndonwords[x], acc)))
        # print(lyndonwords[j],list(map(lambda x: lyndonwords[x], acc)))
  #       if isCompatible(lyndonwords[j],list(map(lambda x: lyndonwords[x], acc))):
            if  visited[j]!= 1:
               acc.append(j)
               visited[j]=1
               
               find_snaker(level+1,j,acc,visited,m,start,neighbors)
            #   print(".")
               visited[j]=0
               found=True
               acc.remove(j)
   if not found :
      found=True
      #print(".",end='')
    #  print(len(acc))
   
      if found:
         if len(acc)>0 and Adjm[acc[0]][acc[-1]]:
            if len(acc)>=mx:
               mx=len(acc)
              # print("**{}****".format(len(acc)))
               #print(acc)   
               print("** ", len(acc), list(map(lambda x: lyndonwords[x], acc)))  
               #print(   "** {} ** {}".format(len(acc), list(map(lambda x: lyndonwords[x], acc))))
         else:
           for j in range(1,len(acc)):
              k= len(acc)-j
#              if Adjm[acc[0]][acc[k]]==1:
              if k>mx and Adjm[acc[0]][acc[k]]:
                mx=k
                # print("* ", k, acc[0:k], list(map(lambda x: lyndonwords[x], acc[0:k])))  
                print("* ", k, lyndonwords[acc[0]],lyndonwords[acc[k]],list(map(lambda x: lyndonwords[x], acc[0:k+1])))  
             #   for x in acc:
             #      print(lyndonwords[x])
             #   list(map(lambda x: print(lyndonwords[x]), acc[0:k+1]))
     #            print( normalizeSequence( transitionSequence(list(map(lambda x: lyndonwords[x], acc[0:k])))))
     #            print(lyndonwords[acc[0]],lyndonwords[acc[k]])
  
              
      #else:
      #   print(len(acc), start, acc.index(cycle))
         #print(len(acc), start, cycle,acc)
      #   print(lyndonwords[cycle],lyndonwords[acc[0]])
               
            
            
def find_snake(level,i,acc,visited,m,start):
   #print(level)
   cycle=None
   found=False
   if i==None:
       for i in range(0,m):
         acc.append(i)
         visited[i]=1
         find_snake(level+1,i,acc,visited,m,i)
         found=True
         visited[i]=0
         acc.remove(i)
   else:     
      for j in range(0,m):

         if  visited[j]!= 1 and Adjm[i][j]==1:
            acc.append(j)
            visited[j]=1
            find_snake(level+1,j,acc,visited,m,start)
            visited[j]=0
            found=True
            acc.remove(j)
   if not found:
      fond=False
      #print(".",end='')
    #  print(len(acc))
      for j in range(0,m):
        if   Adjm[acc[-1]][j]==1: 
           if j==acc[0]:
              found=True
      if found:
         if Adjm[acc[0]][acc[-1]]:
            print(acc)   
            print(   list(map(lambda x: lyndonwords[x], acc)))
            #print(lyndonwords[acc[-1]],lyndonwords[acc[0]])
      #else:
      #   print(len(acc), start, acc.index(cycle))
         #print(len(acc), start, cycle,acc)
      #   print(lyndonwords[cycle],lyndonwords[acc[0]])
               

#
# get Minimal rotation of a word
#
def getMinRotation(word):
   mword=word
   for l in range(1,len(word)):
      x=el(word,l)
      if x < mword:
        mword=x
   return mword
 

#
# check if one word in words list 
# contains a rotation of word
#
def rotavail(word,words):
  #print(word, words)
  word=getMinRotation(word)

  for word2 in words:
    if  word == getMinRotation(word2):
   #    print("True")
       return True  
  #print("False")       
  return False
 
 
solution=[]            

#
# recursive function to search cycles
# level  : recursion level
# sword  : start word
#  word  : actual word
#  l     : list of lyndon words
#  tseq  : transition sequence so far
#  bk    : list of words so far
#  ic    : ćounter hash for transitions of a specific bit position
#
def nextTransition(level,sword, word,l,tseq,bk,ic):
   global solution
   #print("level",level,tseq,word,sword,bk,ic)
   if len(tseq)+1>=l and gray(sword,word):
      tseq.append(transition(word,sword))
      solution=tseq 
      print("Solution",len(tseq), applyTransitionSequence(sword,tseq,[]))
      return True
      
      
   p=[]
   for i in range(0,n):
     p.append(i)
   if tseq != []:
      p.remove( tseq[-1])
   
   random.shuffle( p )
   #print(p)
   for t in p:
  #       print(level,"Transition",t)
         word2=applyTransition(word,t)   
   #      print("Word",word2)
         #if not word2 in bk:
         if not rotavail(word2,bk):
    #      print("not yet there")
         # if word2==sword and len(tseq)>=l:
         #    print("Cycle")
         #    return tseq
     #     print("Lyndon / Necklace",isLyndon(word2),isNecklace(word2))
          if  isNecklace(word2) : # and isLyndon(word2) : #all rotations    
             tseq2 = tseq.copy()
             tseq2.append(t)
             bk2=bk.copy()
             bk2.append(word2)
             if t in ic.keys():
                ic[t] = ic[t] + 1
             else:
                ic[t] = 1
             if nextTransition(level+1,sword,word2,l,tseq2,bk2,ic):
                return True
             #bk2.remove(applyTransition(word2,t))
            # if tseq3 != None:
              #  print("H2")
              #  return tseq3
   return None
 
solutions=[]

def checkNeckalces(words):
# neclace check
 for i in range(0,len(words)):
   word = words[i]
   word0 = word
   for l in range(0,n-1):
      word = e(word)
      if word == word0:
         print("N:",word)
         return False        
 return True
          
  

def checkDuplicates(words):
  # check for duplicates
  flag=False
 
  for i in range(0,len(words)):
    for j in range(0,len(words)):
       if i != j and words[i] == words[j]:
          return True
  return False
  
def compare(l1,l2):
  if len(l1) == len(l2):
    return False
  for i in range(0,len(l1)):
    if l1[i] == l2[i]:
       return False
  return True
  
lmax=0
iterations=0
lastrecord=0
delta=0

#
# Bidirectional Search for Cyclic Gray Code Seed Codes starting from start word
# q         True=>bk1 False => enlarge bk2
# sword     Start Word
# l         desired minimal length 
# bk1       Word list 1
# bk2       Word list 2
#
# 
def bidirectionalNextTransition(level,q,sword,l,bk1,bk2):
   global solution
   global solutions
   global lastrecord
   global iterations
   global lmax
   global delta
 
   iterations = iterations + 1
   
   if bk1 != []:
      word1= bk1[-1]
   if bk2 != []:
      word2= bk2[-1]
      
   # First transition (Bit flip) from same root element (sword) may never be the same in tseq1 and tseq2
   # otherwise it would produce duplicate words
   if len(bk1)==2 and len(bk1)==2 and bk1[1]==bk2[1]:
    #  print("X",tseq1,tseq2)
      return False
  
   # q=True  => Prolongation of tseq1
   # q=False => Prolongation of tseq2
   if  q:
     word=word1
     bk=bk2
   else:
     word=word2
     bk=bk1
    
   
   #if isNecklace(word): # (Method 2) proceed only if word is a necklace (there is no non trivial bit rotation ):    
    #isLyndon(word) and 
   if isLyndon(word) and  isNecklace(word): # (Method 1) proceed only if word is a lyndon word / necklace (there is no non trivial bit rotation ):    
       # both ends meet (if word is next to last element of the "other" list) : we have a cycle 
       # and if lists are sufficient long we record a solution
  #     if  gray(word,bk[-1]) and len(tseq1)+len(tseq2)+1>=l :
        if  gray(word,bk[-1]) and len(bk1)+len(bk2)>=l : 
         bk2.reverse()
         solution = bk1 + bk2
         solution = solution[:-1]
         l = len(solution)
         if l > lmax: 
            lmax=l
            lastrecord=iterations
            
         print("Solution",len(solution),lmax, checkDuplicates(solution),checkNeckalces(solution), solution)         
         solutions.append(solution)

            
         return True
   else:
      return False
      
   #if rotavail(word1,bk2[-1]) or rotavail(word2,bk1[-1]) :
   #      return False
     
   # all possible bit transitions
   p=[]
   for i in range(0,n):
     p.append(i)
   # but never chose same transition position immediately again:
#   if (q and tseq1 != []):
#      p.remove( tseq1[-1])    
#   if (not q and tseq2 != []):
#      p.remove( tseq2[-1])
   
   # randomly chose a transition position:
   random.shuffle( p )
   found=False
   if not q: # prolongate tseq1
     for t in p:
      word1b=applyTransition(word1,t)   
     
      # only if new word is not yet as "rotation" in one of the lists
      #if not rotavail(word1b,bk1) and not  rotavail(word1b,bk2): # (Method 2)
      if not word1b in bk1 and not word1b in bk2: # (Method 1)
         bk1b=bk1.copy()      
         bk1b.append(word1b)
         # try to enlarge:
         if bidirectionalNextTransition(level+1,True, sword,l,bk1b,bk2):
            found=True
            #return True
         if iterations-lastrecord > 5000:
            lastrecord=0
            iterations=0
            delta=len(bk1)
            if delta > level:
              delta = level-1
            print("*****************",delta)
         if delta>0:  
            delta = delta -1
            return found
        
   else: # prolongate tseq2
     for t in p:
      word2b=applyTransition(word2,t)   
      
       # only if new word is not yet as "rotation" in one of the lists
      #if not rotavail(word2b,bk2) and not  rotavail(word2b,bk1): # (Method 2)
      if  not word2b in bk1 and not word2b in bk2: # (Method 1)
        bk2b=bk2.copy()
        bk2b.append(word2b)      
        if bidirectionalNextTransition(level+1,False, sword,l,bk1,bk2b):
            found=True
        if iterations-lastrecord > 5000:
            lastrecord=0
            iterations=0
            delta=len(bk2)
            if delta > level:
              delta = level-1
            print("*****************",delta)
        if delta>0:  
            delta = delta -1
            return found 
           #return True
       
   return found
   
#
# /!\ Work in progress
# Bidirectional Search for Cyclic Gray Code Seed Codes starting from start word
# non recursive version
# q         True=>bk1 False => enlarge bk2
# sword     Start Word
# l         desired minimal length 
# bk1       Word list 1
# bk2       Word list 2
#
# 
def bidirectionalNextTransition2(level,sword,l):
   global solution
   global solutions
   bk1=[sword]
   bk2=[sword]
   q=True
   to_visit=[None,1]
   while(to_visit):
     v = to_visit.pop()
     found=False
     if v :  
        print("bdnt2",v,bk1)
        if q:
         bk=bk1
         obk=bk2
         word=bk[-1]
         nword=applyTransition(word,v)
         if not nword in bk and not nword in obk:
           found=True
           bk.append(nword)
           q=False
         print("bdnt2",v,bk,obk)
        else:
         bk=bk2
         obk=bk1
         word=bk[-1]
         nword=applyTransition(word,v)
         if not nword in bk and not nword in obk:
           found=True
           bk.append(nword)
           q=True
         print("bdnt2'",v,bk,obk)
       #
       # cycle:
       #
        if found:
           if len(bk)+len(obk) >= l and gray(nword,sword) :
             print("Cycle",len(bk)+len(obk),bk,obk)

          # all possible bit transitions
           p=[]
           print("n",n)
           for i in range(0,n):
             p.append(i)
        # randomly chose a transition position:
        #random.shuffle( p )
           print("p",p)
           for x in p:
             to_visit.append(None) # out
             to_visit.append(x) # in
        print("Olga von der Wolga")
     else: # if None we are comming back and pop v
         if bk != []:
             bk.pop()
   return(bk1 + bk2)
     
      
   #if rotavail(word1,bk2[-1]) or rotavail(word2,bk1[-1]) :
   #      return False
     
   # all possible bit transitions
   p=[]
   for i in range(0,n):
     p.append(i)
   # but never chose same transition position immediately again:
#   if (q and tseq1 != []):
#      p.remove( tseq1[-1])    
#   if (not q and tseq2 != []):
#      p.remove( tseq2[-1])
   
   # randomly chose a transition position:
   random.shuffle( p )
   found=False
   if not q: # prolongate tseq1
     for t in p:
      word1b=applyTransition(word1,t)   
     
      # only if new word is not yet as "rotation" in one of the lists
      #if not rotavail(word1b,bk1) and not  rotavail(word1b,bk2): # (Method 2)
      if not word1b in bk1 and not word1b in bk2: # (Method 1)
         bk1b=bk1.copy()      
         bk1b.append(word1b)
         # try to enlarge:
         if bidirectionalNextTransition2(level+1,True, sword,l,bk1b,bk2):
            found=True
            #return True
      
        
   else: # prolongate tseq2
     for t in p:
      word2b=applyTransition(word2,t)   
      
       # only if new word is not yet as "rotation" in one of the lists
      #if not rotavail(word2b,bk2) and not  rotavail(word2b,bk1): # (Method 2)
      if  not word2b in bk1 and not word2b in bk2: # (Method 1)
        bk2b=bk2.copy()
        bk2b.append(word2b)      
        if bidirectionalNextTransition2(level+1,False, sword,l,bk1,bk2b):
            found=True
   
       
   return found
      
# functions to compute cycles from fundamental cycles
#
def getEdgesAt(t,n1):
    acc=[]
    for i in t:
        if i[0]==n1 or i[1]==n1:
            acc.append(i)
    return acc

#
# get path from start node n1 to end node 2 within a tree t
#
# level  : recursion Level 
# t      : edgelist of spanning-tree
# n1,n2  : start and end node
# visited: holds visited nodes
# acc    : accumulator for path
#
def getPathFromTo(level,t,n1,n2,visited,acc):
    if n1==n2:
        return acc
    else:
        l=getEdgesAt(t,n1)
        for e in l:
            if e[1]==n1:
               acc.append(e)
               newvisited = visited.copy()
               newvisited.append(e[0])
               if (not e[0] in visited) and getPathFromTo(level+1,t,e[0],n2,newvisited,acc) != []:
                   return acc
               else:
                   acc.remove(e)
            if e[0]==n1:
               acc.append(e)
               newvisited = visited.copy()
               newvisited.append(e[1])
               if (not e[1] in visited) and getPathFromTo(level+1,t,e[1],n2,newvisited,acc) != []:
                   return acc
               else:
                   acc.remove(e)
    return [] 

#
# check if edges are the same regardless of the direction
#
def isSameEdge(edge1, edge2):
    if (edge1[0] == edge2[0] and edge1[1] == edge2[1]) or (edge1[0] == edge2[1] and edge1[1] == edge2[0] ):
            return True
    return False

#
# check if edges are connected
#
def edgesConnected(edge1, edge2):
    if (edge1[1] == edge2[0] or edge1[0] == edge2[1] or edge1[0] == edge2[0] or edge1[1] == edge2[1]):
            return True
    else:
            return False

#
# check if edge is in list l
#
def edgeInList( edge, l):
    for edge2 in l:
        if isSameEdge(edge,edge2):
            return True
    return False
       
#
# get chords of the spanning tree mst
#   lyndonwords is encoding the nodes of the graph 
#   edge is where nodes have gray distance
#   chords are encoded as hashes
#    h[node index] : list of indices of connected nodes
#
def getChords( lyndonwords, mst ) :
    chords  = {}
    for i in range(0,len(lyndonwords)):
       a = []
       for j in range(0,len(lyndonwords)):
           if i<j and gray(lyndonwords[i],lyndonwords[j]):
              edge = [ i, j, 1]
              if not edgeInList(edge,mst):
                #print("chord" , i,j)
                a.append(j) 
       if a != []:
           chords[i] = a
    return chords

#
# get Edge list of the chords
#
def getEdgeList( chords):
    el = []
    for n1 in chords:
        for n2 in chords  [n1]:
            el.append( [n1,n2,1] )
    return el

#
# normalize the list of edge indices
#
def normalizeEdge( edge ):
    if edge[0] < edge[1]:
        return edge
    else:
        return [ edge[1],edge[0],edge[2] ]

#
# compare function of edges for lexicographic sorting
# [ 2,3,1] < [2,4,1]
#
def edge_compare( e1, e2):
    if e1[0] < e2[0]:
        return -1
    else:
        if e1[0] == e2[0]:
            if e1[1] < e2[1]:
                return -1
            else:
                if e1[1]==e2[1]:
                    return 0
                else:
                    return +1
        else:
            return +1

#
# sort edge list lexicographically
#
def sortEdgeList( chords ):
    #print("chords",len(chords))
    edges = list(map(normalizeEdge,chords))
    edges =sorted(edges,key=itemgetter(0,1))
    #print(edges)
    #print("chords'",len(edges))
    return edges



#
# XOR edge lists -> join lists and remove elements in common
#
def xorEdgeLists(l1, l2):
   x=[]
   for e1 in l1:     
       if not edgeInList(e1, l2):
           x.append(e1)
   for e2 in l2:     
       if not edgeInList(e2 ,l1):
           x.append(e2)
   return x

#
# AND edge lists -> list of common edges of two lists
#
def andEdgeLists(l1, l2):
   x=[]
   for e1 in l1:     
       if not edgeInList(e1,x) and edgeInList(e1, l2):
           x.append(e1)
   for e2 in l2:     
       if not edgeInList(e2,x) and edgeInList(e2 ,l1):
           x.append(e2)
   return x


#
# validate if edge list contains one single cycle
#
def validateSingleCycle( l0 ):
    # lets make a copy because we will successively remove elements:
    l = l0.copy()
    e1 = l.pop()
    while l != []  :
      found=0
      #print("e1=",e1)
      for e2 in l :
        if edgesConnected(e1,e2):
          found=found+1
          if found>1:
              return False
          e1=e2
          l.remove(e2)
      if found==0 and l != []:
          return False

    return True


#
# sequentialize Edge List so that adjacent edges follow each other
# will work only for a cycle
#
def sequentializeEdgeList(l):
    wl= l.copy()
    e1= wl[0]
    sl= []
    sl.append(e1)
    #wl.remove(e1)
    found=True
    while found:
       found=False
       for e2 in wl[1:]:
          if edgesConnected(e1,e2):
             e1=e2
             sl.append(e2) 
             wl.remove(e1)
             found=True
    return sl 

#
# get next connectd edge of an element el in an edge lsit sl
#
def nextConnectedEdge(e1,sl):
    i=1
    wsl=sequentializeEdgeList(sl)
    for e2 in wsl:
        if isSameEdge(e1,e2):
           break
        i=i+1
    for j in range(i,2*len(wsl)):
        k = j % len(wsl)
#        print(e1,wsl[k])
        if edgesConnected(e1,wsl[k]):
            return wsl[k]
    return None
            
#
# generate node list from edge list
#   words : node words 
#   sl    : edge list
#
def getNodeList(words,sl):
    # first edge:
    e0= sl[0]
    # actual edge initialized with first one:
    e1= sl[0]
    # visited edge list:
    v = []
    # node list:
    nl=[words[e1[0]]]
    # tail end:
    q=0
    loop=True
    while loop:
       # get next edge:
       e2=nextConnectedEdge(e1,sl)
       if not e2 in v:
         if q==0 :
           if e1[1]==e2[0]:
             nl.append( words[e2[0]] )
             q=0
           else:
             if e1[1]==e2[1]:
               nl.append( words[e2[1]] )
               q=1
         else:
             if e1[0]==e2[0]:
               nl.append( words[e2[0]] )
               q=0
             else:
               if e1[0]==e2[1]:
                  nl.append( words[e2[1]] )
                  q=1
         e1=e2
         v.append(e2)
       else:
         loop=False

       # Terminate if no edge is found or edge is first one
       if e2==None or isSameEdge(e0,e2):
#           print("Cycle complete")
           loop=False
           
       e1=e2
    return nl 


#
# validate cycle and print it if valid
#
h = {}
xmax=0
x=0
def validateAndPrintCycle(l):
    global h
    global xmax
    global x
    valid=validateSingleCycle(l)
    if valid:
      if len(l) in h.keys():
        h[len(l)] = h[len(l)] + 1
      else:
        h[len(l)]=1
        x = len(l)
        if x > xmax:
          xmax=len(l)
          print("SingleCycle ({}) {} {} ".format(len(l),valid,l))


def edgeIndexing(el):
   idxToEdge= []
   edgeToIdx = {} 
   i=0
   for e in el:
      if e[0] < e[1]:
        et = tuple( [e[0],e[1]])
      else:
        et = tuple( [e[1],e[0]])
      if not et in idxToEdge:
          idxToEdge.append(et)
          edgeToIdx[et] = i
          i=i+1
   return (idxToEdge,edgeToIdx)

def normalizedEdge(e):
    if e[0] < e[1]:
        return  [e[0],e[1],1]
    else:
        return  [e[1],e[0],1]

def edgesToIndex(el,eidx):
    a=[]
    for i in range(0,len(eidx)):
       en = normalizedEdge(eidx[i])
       if en in el:
          a.append(1)
       else:
          a.append(0)
    return a

#
# class for distance cycle matrix
#
class DistanceMatrix:

    def __init__(self,cycs):
          self.cycles=cycs
          self.cycleLens = []
          self.cycleOverlap = []
          self.combinations = set([])

#          for c in self.cycles:
#              self.cycleLens.append(len(c))
          for i in range(0,len(self.cycles)):
             self.cycleLens.append(len(self.cycles[i]))
             row=[]
             for j in range(0,len(self.cycles)):
                 if i==j:
                   row.append(0)
                 else:
                   l=andEdgeLists(self.cycles[i],self.cycles[j])
                   row.append(len(l))
             self.cycleOverlap.append(row) 

    def addCombination(self,i,j):
        self.combinations.add( tuple([i,j]) )
    
    def hasCombination(self,i,j):
        if tuple([i,j]) in self.combinations:
            return True
        else:
            return False
        
    def getLen(self):
        return len(self.cycles)

    def getLens(self,i,j):
         #print(i,j,len(self.cycleLens),len(self.cycleOverlap[i]))
         return tuple( [self.cycleLens[i], self.cycleLens[j],self.cycleOverlap[i][j]])

    def add(self,cyc):
#        print("LC",len(self.cycles))
        self.cycleLens.append(len(cyc))
        self.cycles.append(cyc)
        for i in range(0,len(self.cycles)-1):
            l=andEdgeLists(cyc,self.cycles[i])
            self.cycleOverlap[i].append(len(l))

        row=[]
        for j in range(0,len(self.cycles)):
          if j==len(self.cycles):
            row.append(0)
          else:
            l=andEdgeLists(cyc,self.cycles[j])
            row.append(len(l))

        self.cycleOverlap.append(row) 

    def getCycleCombinations(self,n,m):
       ccl = []
       for i in range(0,len(self.cycles)):
         for j in range(i+1,len(self.cycles)):
             if not self.hasCombination(i,j): # if combination has been given last time skip it 
               x=self.getLens(i,j)
               if x[2] != 0: # no double cycles
                 nl = x[0] + x[1] - 2*x[2]
                 y= [i,j, nl ,x[0],x[1],x[2]]
             
               #if nl<=n:
                 self.combinations.add( tuple([i,j]) )
                 ccl.append(y) 
       ccl=sorted(ccl,reverse=True,key=itemgetter(2))      
       print("Cycles:" ,len(self.cycles),k)
       if m!=0:
           ccl = ccl[0:m]
       return ccl

          
def printBitList(bitlist):
    for bit in bitlist:
        print(bit,end=' ')
    print("")

#
# DFS combination of fundamental cycles
#  (fundamental cycles are basis of all cycles of a graph)
#
#  fcs    : Array with fundamental cycles
#  a      : Bit-Matrix
#  idx    : list of indices of already combined fundamental cycles
#  cycle  : Edge list containing current cycle
#
lmax=0
def searchFundamentalCycleCombinations(fcs,a,idx,cycle,lns):
  global lmax

  p=[]
  for x in range(0,len(fcs)):
     p.append(x)
  random.shuffle(p)

  for i in p:               # try all fundamental cycles (randomized order)
     if i not in idx:       # use each cycle only once
        idx.append(i)       # mark cycle index so that it will not be used again
        fc=fcs[i]
        l=xorEdgeLists(cycle ,fc) # combine current cycle with i-th  fundamental cycle 

        if l != [] and validateSingleCycle(l):
          if len(l)>lmax:
             lmax=len(l)
             print("Found", len(l),lmax,l)
             print("Cycles combined:", idx)
             sl=sequentializeEdgeList(l)
             print("A",len(sl),getNodeList( lns,sl))

    #         ix0=idx[0]
    #         c=a[ix0]
             #print(ix0,c)
    #         for ix in idx[1:]:
    #             c=xorCycles(c,a[ix])
        #         print(ix,a[ix])
        #         print("::",c)
                 #printBitList(a[ix])

#             l1=[]
#             for j in range(1,len(idx)):
#                 l2 = fcs[idx[j]]
#                 a12=andEdgeLists(l1,l2)
#                 print(j,idx[j],l1,l2,a12)
#                 l1=xorEdgeLists(l1,l2)
#                 print(l1,l2)
#                 
#             for ix in idx:
#                 print(fcs[ix])
          searchFundamentalCycleCombinations(fcs,a,idx,l,lns)
        idx.remove(i)


def sum(a):
    s=0
    for x in a:
        s = s+ x
    return s

def andCycles(c1,c2):
 c12 = []
 for i in range(0,len(c1)):
     if c1[i]==1 and c2[i]==1:
         c12.append(1)
     else:
         c12.append(0)
 return c12

def xorCycles(c1,c2):
 c12 = []
 for i in range(0,len(c1)):
    c12.append((c1[i] + c2[i]) % 2)
 return c12

#
#
#
lmax=0
def combine(level,a,c,idxToEdge,idx):
  global lmax
#  print(level,sum(c),c)

  s=[]
  for i in range(0,len(c)):
      if c[i]==1:
          s.append( idxToEdge[i])

  if s!=[] and validateSingleCycle(s):
    if len(s)>lmax:
        lmax=len(s)
        print("S=",len(s),lmax,s)
        print("IDX:",idx)
  else:
    return

  loop=False
  if a==[]:
      return

  p=[]
  for k in range(0,len(a)):
      p.append(k)

  #random.shuffle(p)
  
  for k in p:
    l=andCycles(c,a[k])
    if sum(l) != 0:
        loop=True
      
    if not loop:
      return

    cn=xorCycles(c,a[k])
    state=0
    an=[]
    for i in range(0,len(a)):
      if not i==k:
         l2=andCycles(l,a[i])
         if sum(andCycles(l,l2)) <  1:
           an.append(a[i])
           state=1
    if state==1:           
       combine(level+1,an,cn,idxToEdge,idx) 

    
#
# Calculataion via fundamental cycles:
#
# https://www.codeproject.com/Articles/1158232/Enumerating-All-Cycles-in-an-Undirected-Graph
#
def fundamentalCycleConstruction(lns):
#
# Step 1: Generate Graph from Lyndon-Necklaces:
#
#print( "LNS:", lns)
#print( len( lns) ) 
  print("Step 1: Generate Graph from Lyndon Necklaces:")
  g= graph.Graph(len(lns))
  for i in range(0,len(lns)):
    for j in range(0,len(lns)):
      if gray(lns[i],lns[j]):
#        print("edge",i,j)
        g.addEdge(i,j,1)

#
# Step 2: Compute spanning tree of graph via Kruskal's algorithm
#

  print("Step 2: Generate Graph Spanning Tree:")
  print(g.graph)
  t=g.KruskalMST()
  print(t)

#
# Step 3: Get all Chords of the graph (edges not in spanning tree)
#
  print("Step 3: Chords")
  cs = getChords(lns,t)
  cl = getEdgeList(cs)

  print("Vertices:",(g.V))
  print("Edges:",int(len(g.graph) / 2))
  print("Chords:",len(cl))
  print("Edges - Vertices + 1:",int(len(g.graph)/2 - g.V + 1))
  for c in cs:
    print(c,cs[c])

  nl=[]
  for e in g.graph:
      if not e[0] in nl:
          nl.append(e[0])
      if not e[1] in nl:
          nl.append(e[1])
  print(nl) 
  for n in nl:
    e = []
    for x in  g.graph:
      if x[0]==n and not x[1] in e:
          e.append(x[1])
      if x[1]==n and not x[0] in e:
          e.append(x[0])
    print(n,e)
    

  (idx2edge,edge2idx) = edgeIndexing(g.graph)
   


#
# Step 4: Compute all Fundamental Cycles of the graph
# one fundamental circle per chord
#

  print("Step 4: Calculating fundamental cycles")
  cl = getEdgeList(cs)
  fcs=[]
  lfcs=[]
  for i in range(0,len(cl)):
    c = cl[i]
    fc = sortEdgeList(getPathFromTo(0,t,c[0],c[1],[c[0]],[]))
    fc.append(c)
    fcs.append(fc)
    lfcs.append(  (i,len(fc)))


  slfcs = sorted(lfcs,reverse=True, key=lambda x : x[1])
  # (cycle index, length)
  print(slfcs)

#  print("Fundamental cycles:")
  i=0
  alist=[]
  for fc in fcs:
    a=edgesToIndex(fc,idx2edge)
    print("a",i,a)
    alist.append(a)
    i=i+1
#    print("Fundamental cycle:",fc)
  
#  for i in range(0,len(alist)):
#  combine(0,alist,alist[2],idx2edge,[2])

  print("Step 5: Combine Cycles:")
  searchFundamentalCycleCombinations(fcs,alist,[],[],lns)

#  l=xorEdgeLists(fcs[0] ,fcs[1])
#  l=xorEdgeLists(l ,fcs[2])
#  l=xorEdgeLists(l ,fcs[4])

#  if validateSingleCycle(l):
#     l=sequentializeEdgeList(l)
#     print("A",len(l),getNodeList( lns,l))
#     print("C",validateSingleCycle(l),l)
#  else:
#     print("Not valid")


  o=9/0
  llfcs=len(fcs)

  dm = DistanceMatrix(fcs)
  print("Pass 1")
  ccs=dm.getCycleCombinations(30,len(fcs))
  for c in ccs:
      l=xorEdgeLists(fcs[c[0]] ,fcs[c[1]])
      print(c[0],fcs[c[0]])
      print(c[1],fcs[c[1]])
      print(l)
      dm.add(l)

#  print("Pass 2")
#  ccs=dm.getCycleCombinations(30,200)
#  print("ccs",len(ccs))
#  for c in ccs:
#      l=xorEdgeLists(fcs[c[0]] ,fcs[c[1]])
#      dm.add(l)

#  print("Pass 3")
#  ccs=dm.getCycleCombinations(20,200)
#  print("ccs",len(ccs))
#  for c in ccs:
#      l=xorEdgeLists(fcs[c[0]] ,fcs[c[1]])
#      dm.add(l)

#  print("Pass 4")
#  ccs=dm.getCycleCombinations(20,85)
#  for c in ccs:
#      l=xorEdgeLists(fcs[c[0]] ,fcs[c[1]])
#      dm.add(l)


  sols=dm.cycles

  #lmax=0
  #for i in range(0,dm.getLen()):
  #    for j in range(i+1,dm.getLen()):
  #       x=dm.getLens(i,j)
  #       if x[2] != 0: # only single cycles!
  #          lact= x[0]+x[1]-2*x[2] 
  #          sols.append( [   lact , i,j ] )
  #          if lact > lmax:
  #              lmax=lact 
  #  sols=sorted(sols,reverse=True,key=itemgetter(0))

#  print( "LMAX",lmax)
  lmax=0
  for l in sols:
     #l=xorEdgeLists( dm.cycles[ sol[1] ], dm.cycles[ sol[2] ])
     if len(l)>5:
       if validateSingleCycle(l):
          l=sequentializeEdgeList(l)
          print("A",len(l),getNodeList( lns,l))
          print("C",validateSingleCycle(l),l)
          #print("D",lns)
          if len(l)>lmax:
             lmax=len(l)
  print( "LMAX",lmax)

  o=9/0
#
# Step 5: Combine Fundamental Cycles via XOR-Operation
#         and validate if results are single cycles
#
  loop=4
  p1=0
  p2=0
  lmax=0
  while loop>0:
   p1=0 # p2
   p2=len(slfcs)
   pmax=p2
   for i1 in range(p1,p2):
    for i2 in range(i1+1,p2):
     i3 = p2 - i2
     if i1 != i3:
      l=xorEdgeLists(fcs[slfcs[i1][0]],fcs[slfcs[i3][0]])
      if l != [] and validateSingleCycle(l):
         if len(l)>lmax:
             lmax=len(l)
         if pmax % 1000 == 0:
            print("GOGO",len(l),lmax)
         fcs.append(l)
         slfcs.append( (pmax,len(l)))
         pmax = pmax + 1
     
   slfcs = sorted(slfcs,reverse=True, key=lambda x : x[1])
   print("SLFCS:",slfcs)
   loop = loop - 1

#    for i2 in range(i1+1,lfcs-1):
#       validateAndPrintCycle(l)
#       if validateSingleCycle(l):
#        validateAndPrintCycle(l)
#        for i3 in range(i2+1,lfcs-1):
#         l=xorEdgeLists(l,fcs[i3])
#         if validateSingleCycle(l):
#          validateAndPrintCycle(l)
#          for i4 in range(i3+1,lfcs-1):
#           l=xorEdgeLists(l,fcs[i4])
#           validateAndPrintCycle(l)


#l=sequentializeEdgeList(0,l,[])
#
# Print statistics of found Cycle lengths
#

  for k in h:
      print(k,h[k])


#
# uni- and bidirectional DFS-Search Construction
#
#  n      : word length                      (1st program arg)
#  seqlen : desired minimal sequence length  (2nd program arg)
#
def bidirectionalSearchConstruction(n,seqlen):
  print("Knallkoff")

  s0=""
  for i in range(0,n-1):
    s0= s0+"0"
  s0=s0+"1"

# unidirectional Search
#  solution=nextTransition(0,s0, s0,seqlen,[],[s0],{} )

# (ok)    :
  bidirectionalNextTransition(0,True,s0,seqlen,[s0],[s0])
# (not ok):   
#  solution= bidirectionalNextTransition2(0,s0,seqlen)
#  print( "Solution", len(solution),solution)

  print(solutions)
  print(len(solutions))

#
# snake search construction
# lns:      Lnyon words / Necklaces
#
def snakeSearchConstruction(lns):
  nl= len(lns)
  # Create Adjacence Matrix of lyndon words
  initAdjm(nl)
  fillAdjm(lyndonwords,nl)
#o=9/0
  printAdjm(nl)
  find_neighbors(lyndonwords)
  (mi,ma)=split_min_max()
  print(mi,ma)
  mins=mi
  maxs=ma

  visited=[]
  for i in range(0,nl):
    visited.append(0)
  Tc=0;  
  print("Hurra!")

  # find the start word - 0...01
  s=""
  for i in range(0,n-1):
    s= s+"0"
  s=s+"1"
  s=lyndonwords.index(s)


  # recursive search for snakes / cycles:
  s=find_snaker(0,s ,[s],visited,nl,None,neighbors)

  print(sort_gray( 0, lyndonwords[1::], [lyndonwords[0]] ))
#print(transitionSequence(s))




#
# Compute Lyndon Words of given length n:
#

# To store the indices
# of the characters
w = [-1]
  
# Loop till w is not empty
while w:
    print ("A",w)  
    # Incrementing the last character
    w[-1] += 1
    m = len(w)
    print("B",w)
    if m == n:
        print("C",w)
        #print(''.join(S[i] for i in w))
        lyndonwords.append(''.join(S[i] for i in w))
    # Repeating w to get a
    # n-length string
    while len(w) < n:
        w.append(w[-m])
    
    # Removing the last character
    # as long it is equal to
    # the largest character in S
    while w and w[-1] == k - 1:
        w.pop()
 
#s =  [1, 0, 2, 3, 7, 5, 4, 6, 14, 10, 8, 9, 11, 15, 13, 12, 26, 19, 20, 22, 49, 54, 42, 41, 21, 17, 24, 43, 34, 35, 33, 36, 45, 47, 44, 25, 18] 
#s =  [1, 0, 2, 3, 7, 6, 4, 5, 14, 15, 17, 16, 10, 9, 12, 13, 11, 8]
#print(   list(map(lambda x: lyndonwords[x], s)))
#o=9/0
print(  lyndonwords)
print( "Lyndonwords:", len(lyndonwords))

#
# Only Lyndon Words that are also Necklaces:
#
lns= []
for x in lyndonwords:
    if isNecklace(x):
        lns.append(x)

# uncomment 1 Method to be used for search:

#fundamentalCycleConstruction(lns)

# very fast
bidirectionalSearchConstruction(n,seqlen)

#snakeSearchConstruction(lns)
